import RoboidControl.Messages.LowLevelMessages as LowLevelMessages
from RoboidControl.Messages.IMessage import IMessage
from RoboidControl.Thing import Thing
from LinearAlgebra.Spherical import Spherical
from LinearAlgebra.Quaternion import Quaternion

from typing import Union, Optional
import logging
logger = logging.getLogger(__name__)

class PoseMsg(IMessage):
    """! Message to communicate the pose of the thing
    The pose is in local space relative to the parent. If there is no parent
    (the thing is a root thing), the pose will be in world space.
    """
    ## The message ID
    id = 0x10
    ## The length of the message in bytes
    length = 4

    ## Bit pattern for a pose with position
    Pose_Position = 0x01
    ## Bit pattern for a pose with orientation
    Pose_Orientation = 0x02    
    ## Bit pattern for a pose with linear velocity
    Pose_LinearVelocity = 0x04
    ## Bit pattern for a pose with angular velocity
    Pose_AngularVelocity = 0x08

    def __init__(self, arg1: Union[bytes, int], thing: Optional[Thing], force: bool = False):
        """! Create a PoseMsg
        @param arg1 Either:
        - bytes: contianing the serialized data; or
        - int: containgint the network_id 
        @param thing The thing for which the data is communicated
        @param force If true, position and orientation are always included, even when they are not updated
        """

        if thing is None:
            buffer = bytes(arg1)
            ## The network ID of the thing
            self.network_id = buffer[1]
            ## The ID of the thing
            self.thing_id = buffer[2]
            ## Bit pattern stating which pose components are available
            self.pose_type = buffer[3]

            ## The position of the thing in local space in meters
            self.position: Optional[Spherical] = None
            ## The orientation of the thing in local space
            self.orientation: Optional[Quaternion] = None
            ## The linear velocity of the thing in local space in meters per second
            self.linear_velocity: Optional[Spherical] = None
            ## The angular velocity of the thing in local space
            self.angular_velocity: Optional[Spherical] = None

            ix:int = 4
            ix_ref = [ix]
            if self.pose_type & PoseMsg.Pose_Position != 0:
                self.position = LowLevelMessages.ReceiveSpherical(buffer, ix_ref)
            if self.pose_type & PoseMsg.Pose_Orientation != 0:
                self.orientation = LowLevelMessages.ReceiveSwingTwist(buffer, ix_ref)
            if self.pose_type & PoseMsg.Pose_LinearVelocity != 0:
                self.linear_velocity = LowLevelMessages.ReceiveSpherical(buffer, ix_ref)
            if self.pose_type & PoseMsg.Pose_AngularVelocity != 0:
                self.angular_velocity = LowLevelMessages.ReceiveSpherical(buffer, ix_ref)
        else:
            ## The network ID of the thing
            self.network_id = arg1
            ## The ID of the thing
            self.thing = thing

            ## Bit pattern stating which pose components are available
            self.pose_type = 0
            if thing.position_updated or force:
                self.pose_type |= PoseMsg.Pose_Position
            if thing.orientation_updated or force:
                self.pose_type |= PoseMsg.Pose_Orientation
            if thing.linear_velocity_updated or force:
                self.pose_type |= PoseMsg.Pose_LinearVelocity
            if thing.angular_velocity_updated or force:
                self.pose_type |= PoseMsg.Pose_AngularVelocity

    def Serialize(self, buffer_ref):
        """!
        @copydoc RoboidControl::Messages::IMessage::IMessage::Serialize
        """ 
        if self.thing is None or self.pose_type == 0:
            return 0
        
        logger.debug(f'Send PoseMsg [{self.network_id}/{self.thing.id}] {self.pose_type}')

        buffer: bytearray = buffer_ref[0]
        buffer[0:PoseMsg.length] = [
            PoseMsg.id,
            self.network_id,
            self.thing.id,
            self.pose_type
        ]
        ix = [4]
        if self.pose_type & PoseMsg.Pose_Position:
            LowLevelMessages.SendSpherical(buffer, ix, self.thing.position)
        if self.pose_type & PoseMsg.Pose_Orientation:
            LowLevelMessages.SendSwingTwist(buffer, ix, self.thing.orientation)
        if self.pose_type & PoseMsg.Pose_LinearVelocity:
            LowLevelMessages.SendSpherical(buffer, ix, self.thing.linear_velocity)
        if self.pose_type & PoseMsg.Pose_AngularVelocity:
            LowLevelMessages.SendSpherical(buffer, ix, self.thing.angular_velocity)
        return PoseMsg.length + ix[0]
